#include "global.h"
#include "print.h"
#include "modeman.h"
#include "bddman.h"
#include "zoneman.h"
#include "back.h"
#include "reduce.h"

struct a23tree_type *mode_mu_tree;

int back_reduce(struct thread_type *t);
int back_path(struct mode_type *, struct transition_type *);
int consistent_inv_mu(struct mode_type *, struct clkconj_type *, bdd);
int back_dfs(struct mode_mu_type *);
int mode_mu_comp(struct mode_mu_type *, struct mode_mu_type *);
int free_mode_mu(struct mode_mu_type *);
int mode_mu_count_inc(struct mode_mu_type *);
int mode_mu_parent_set(struct mode_mu_type *, struct a23tree_type *);
int print_mode_mu(struct mode_mu_type *);
struct clkconj_type *back_delta(struct clkconj_type *, int);
int initial_consistent(struct mode_mu_type *);
struct clkconj_type *reset_clk_assign(struct clkconj_type *, 
                                      struct assign_type *);
bdd reset_prp_assign(bdd, struct assign_type *);
struct clkconj_type *weakest_clk_precondition(struct transition_type *,
                                              struct mode_mu_type *,
                                              struct clkconj_type *);
bdd weakest_prp_precondition(struct transition_type *, struct mode_mu_type *);

/**** FUNCTIONS ****/

int back_reduce(struct thread_type *t) {
  struct mode_type *M;
  struct transition_type *e, *se;

  printf("Reducing thread %d using backward search.\n", t->index);
  if(verbose >= 0) {
    sprintf(tmpstring,
            "\nReducing thread %d using backward search:\n", t->index);
    output(tmpstring);
  }

  for(M=t->starting_mode; M; M=M->cnext) {
    if(M->out_trans_cycle) {
      for(e=M->out_trans_cycle->next, se=e->next; ; e=se, se=se->next) {
        init_23tree(&(mode_mu_tree));
        if(back_path(M, e) == FALSE){          /* will never become valid */

          eliminate_transition(e, M);

          sprintf(tmpstring, 
                  "ELIMINATED transition %d from mode %s (%d) to %s (%d).\n",
                  e->index, M->name, M->index, e->mode->name, e->mode->index);
          output(tmpstring);

          t->trans_count--;
          t->invalid_trans_count--;
          if(M->out_trans_cycle == NULL || se==M->out_trans_cycle->next) break;

        }
        if(e==M->out_trans_cycle) break;
      }
    }
  }
  printf("\tRemoving irreachable modes in thread %d.\n", t->index);
  remove_irreachable_mode(t);
  print_counts(t->index);
}

int back_path(struct mode_type *M, struct transition_type *e) {
  bdd mu_prp, tmp_bdd;
  struct clkconj_type *mu_clk;
  struct mode_type *m, *Mm;
  struct mode_mu_type *x, *nx;

  sprintf(tmpstring, 
          "\nConsidering mode %s (%d) and transition %d to mode %s (%d).\n", 
          M->name, M->index, e->index, e->mode->name, e->mode->index);
  output(tmpstring);

  mu_prp = BDD_AND(M->boolexp->boolexp_bdd, e->trigger_bdd);

  tmp_bdd = BDD_AND(mu_prp, M->dval_bdd);
  if(tmp_bdd != BDD_FALSE) {
    bdd_free(bddm, tmp_bdd);
    output("\tA valid transition.\n");
    return TRUE;
  }
  bdd_free(bddm, tmp_bdd);

  mu_clk = clkconj_and(M->xconj_cycle, e->clk_trigger_cycle);
  reset_child_visits(M);

  for(m = thread[total_automata-1].starting_mode; m; m = m->cnext) {

    sprintf(tmpstring, 
            "    Considering A^1 mode %s (%d).\n", m->name, m->index);
    output(tmpstring);

    if(consistent_inv_mu(m, mu_clk, mu_prp)) {

      output("\t... consistent mode.invariance and mu.\n");

      Mm = compose_modes(M, m);
      
      x = (struct mode_mu_type *)malloc(sizeof(struct mode_mu_type));
      x->mode = Mm;
      x->mu_clk = mu_clk;
      x->mu_prp = mu_prp;
      x->visited = FALSE;

/*
      nx = (struct mode_mu_type *)
            add_23tree(x, mode_mu_tree, mode_mu_comp,
            free_mode_mu, mode_mu_count_inc, mode_mu_parent_set, print_mode_mu);

      if(nx == x) {
        if(nx->mode->submode1->child_cycle) {
          nx->next_child = nx->mode->submode1->child_cycle->next_child;
          nx->mode->submode1->child_cycle->next_child = nx;
          nx->mode->submode1->child_cycle = nx;
        }
        else x->mode->submode1->child_cycle = x->next_child = x;
      }
*/

      if(back_dfs(x) == TRUE) {
        return TRUE;
      }
    }
    else 
      output("\t... inconsistent mode.invariance and mu.\n");
  }
  return FALSE;
}

int consistent_inv_mu(struct mode_type *m, struct clkconj_type *mu_clk,
                      bdd mu_prp) {
  struct clkconj_type *c;
  bdd b;

  c = clkconj_and(m->xconj_cycle, mu_clk);
  if(c->op == FALSE) {
    free_clkconj(c);
    return FALSE;
  }
  free_clkconj(c);

  b = BDD_AND(m->boolexp->boolexp_bdd, mu_prp);
  if(b == BDD_FALSE) {
    bdd_free(bddm, b);
    return FALSE;
  }
  bdd_free(bddm, b);

  return TRUE;
}

int back_dfs(struct mode_mu_type *x) {
  struct clkconj_type *mu1, *mu_clk_before, *mu_clk_after, *tmp_clkconj;
  bdd mu_prp_before, mu_prp_after;
  struct transition_type *f;
  struct mode_type *Mm, *Mn, 
                   *M = x->mode->submode1, 
                   *m = x->mode->submode2, 
                   *n, *nm;
  struct mode_mu_type *y, *ny, *z, *nz;


  sprintf(tmpstring, "\tBackward searching from (%s, %s) with:\n", 
          M->name, m->name);
  output(tmpstring);
  
  output("\tclock constraint: ");
  print_clkconj(x->mu_clk);
  output("\n");
  output("\tbdd: ");
  print_bdd(x->mu_prp);

  if(x->visited == TRUE) {
    output("\tVISITED!\n");
    return FALSE;
  }
  else if(initial_consistent(x)) {
    output("\tINITIAL modes reached and consistent.\n");
    return TRUE;
  }
  else {

    mu1 = back_delta(x->mu_clk, FALSE);
    mu1 = conjoin(mu1, DELTA);

    if(m->in_trans_cycle) {
      for(f=m->in_trans_cycle->next; ; f=f->next) {

        sprintf(tmpstring, "\t  considering in-trans %d from A^1 mode %s.\n",
                f->index, f->mode->name);
        output(tmpstring);

        mu_clk_before = weakest_clk_precondition(f, x, mu1);
        mu_prp_before = weakest_prp_precondition(f, x);

        output("\t  mu_clk_before: "); print_clkconj(mu_clk_before);
        output("\n");
        output("\t  mu_prp_before: "); print_bdd(mu_prp_before);
/*
        output("Assigning: "); print_assignment1(f->assign_cycle); output("\n");
        output("TO: "); print_clkconj(mu_clk_before); output("\n");
*/
        tmp_clkconj = clkconj_copy(mu_clk_before);
        mu_clk_after = assign(tmp_clkconj, f->assign_cycle);
/*
        output("After assign(): "); print_clkconj(mu_clk_after); output("\n");
*/
        tmp_clkconj = clkconj_and(mu_clk_after, mu1);
        free_clkconj(mu_clk_after);
        mu_clk_after = tmp_clkconj;

        tmp_clkconj = clkconj_and(mu_clk_after, x->mu_clk);
        if(tmp_clkconj->op == FALSE) {
          free_clkconj(tmp_clkconj);
          output("mu_after is inconsistent with mu (CLK).\n");
          if(f==m->in_trans_cycle) break;
          else continue;
        }
        free_clkconj(tmp_clkconj);

        mu_prp_after = bdd_assign(mu_prp_before, f->assign_cycle);

        if(BDD_AND(mu_prp_after, x->mu_prp) == BDD_FALSE) {
          output("mu_after is inconsistent with mu (BDD).\n");
          if(f==m->in_trans_cycle) break;
          else continue;
        }

        output("\t  mu_clk_after: "); print_clkconj(mu_clk_after);
        output("\n");
        output("\t  mu_prp_after: "); print_bdd(mu_prp_after);

        Mm = compose_modes(M, m);

        y = (struct mode_mu_type *)malloc(sizeof(struct mode_mu_type));
        y->mode = Mm;
        y->mu_clk = mu_clk_after;
        y->mu_prp = mu_prp_after;
        y->visited = TRUE;

        ny = (struct mode_mu_type *)
              add_23tree(y, mode_mu_tree, mode_mu_comp, free_mode_mu, 
                         mode_mu_count_inc, mode_mu_parent_set, print_mode_mu);

        if(ny == y) {
          output("PUSHED mode_mu.\n");
          if(y->mode->submode1->child_cycle) {
            y->next_child = y->mode->submode1->child_cycle->next_child;
            y->mode->submode1->child_cycle->next_child = y;
            y->mode->submode1->child_cycle = y;
          }
          else y->mode->submode1->child_cycle = y->next_child = y;
        }
        else {
          output("FOUND similar mode_mu.\n");
          if(f==m->in_trans_cycle) break;
          else continue;
        }

        for(n = thread[total_automata-1].starting_mode; n; n = n->cnext) {
          if(consistent_inv_mu(n, mu_clk_before, mu_prp_before)) {
          
            sprintf(tmpstring, 
                    "\t    considering next consistent A^1 mode %s (%d).\n", 
                    n->name, n->index);
            output(tmpstring);

            Mn = compose_modes(M, n);

            z = (struct mode_mu_type *)malloc(sizeof(struct mode_mu_type));
            z->mode = Mn;
            z->mu_clk = mu_clk_before;
            z->mu_prp = mu_prp_before;
            z->visited = FALSE;

/*
            nz = (struct mode_mu_type *)
                 add_23tree(z, mode_mu_tree, mode_mu_comp, free_mode_mu, 
                            mode_mu_count_inc, mode_mu_parent_set, 
                            print_mode_mu);

            if(nz == z) {
              output("PUSHED mode_mu.\n");
              if(z->mode->submode1->child_cycle) {
                z->next_child = z->mode->submode1->child_cycle->next_child;
                z->mode->submode1->child_cycle->next_child = z;
                z->mode->submode1->child_cycle = z;
              }
              else z->mode->submode1->child_cycle = z->next_child = z;
            }
            else output("FOUND similar mode_mu.\n"); 
*/
 
            if(back_dfs(z) == TRUE) {
              return TRUE;
            }
          }
        } /* end for(n = ...) */
        if(f==m->in_trans_cycle) break;
      }
    }


    if(M->in_trans_cycle) {
      for(f=M->in_trans_cycle->next; ; f=f->next) {

        sprintf(tmpstring, "\t  considering in-trans %d from A^k mode %s.\n",
                f->index, f->mode->name);
        output(tmpstring);

        mu_clk_before = weakest_clk_precondition(f, x, mu1);
        mu_prp_before = weakest_prp_precondition(f, x);

        output("\t  mu_clk_before: "); print_clkconj(mu_clk_before);
        output("\n");
        output("\t  mu_prp_before: "); print_bdd(mu_prp_before);

        tmp_clkconj = clkconj_copy(mu_clk_before);
        mu_clk_after = assign(tmp_clkconj, f->assign_cycle);

        tmp_clkconj = clkconj_and(mu_clk_after, mu1);
        free_clkconj(mu_clk_after);
        mu_clk_after = tmp_clkconj;

        tmp_clkconj = clkconj_and(mu_clk_after, x->mu_clk);
        if(tmp_clkconj->op == FALSE) {
          free_clkconj(tmp_clkconj);
          output("mu_after is inconsistent with mu (CLK).\n");
          if(f==M->in_trans_cycle) break;
          else continue;
        }
        free_clkconj(tmp_clkconj);

        mu_prp_after = bdd_assign(mu_prp_before, f->assign_cycle);

        if(BDD_AND(mu_prp_after, x->mu_prp) == BDD_FALSE) {
          output("mu_after is inconsistent with mu (BDD).\n");
          if(f==M->in_trans_cycle) break;
          else continue;
        }

        output("\t  mu_clk_after: "); print_clkconj(mu_clk_after);
        output("\n");
        output("\t  mu_prp_after: "); print_bdd(mu_prp_after);

        Mm = compose_modes(M, m);

        y = (struct mode_mu_type *)malloc(sizeof(struct mode_mu_type));
        y->mode = Mm;
        y->mu_clk = mu_clk_after;
        y->mu_prp = mu_prp_after;
        y->visited = TRUE;

        ny = (struct mode_mu_type *)
              add_23tree(y, mode_mu_tree, mode_mu_comp, free_mode_mu, 
                         mode_mu_count_inc, mode_mu_parent_set, print_mode_mu);

        if(ny == y) {
          output("PUSHED mode_mu.\n");
          if(y->mode->submode1->child_cycle) {
            y->next_child = y->mode->submode1->child_cycle->next_child;
            y->mode->submode1->child_cycle->next_child = y;
            y->mode->submode1->child_cycle = y;
          }
          else y->mode->submode1->child_cycle = y->next_child = y;
        }
        else {
          output("FOUND similar mode_mu.\n");
          if(f==M->in_trans_cycle) break;
          else continue;
        }

        if(consistent_inv_mu(f->mode, mu_clk_before, mu_prp_before)) {

          output("\t    In-trans-mode is consistent with mu_before.\n");

          nm = compose_modes(f->mode, m);
          z = (struct mode_mu_type *)malloc(sizeof(struct mode_mu_type));
          z->mode = nm;
          z->mu_clk = mu_clk_before;
          z->mu_prp = mu_prp_before;
          z->visited = FALSE;

/*
          nz = (struct mode_mu_type *)
               add_23tree(z, mode_mu_tree, mode_mu_comp, free_mode_mu, 
                          mode_mu_count_inc, mode_mu_parent_set, 
                          print_mode_mu);

          if(nz == z) {
            output("PUSHED mode_mu.\n");
            if(z->mode->submode1->child_cycle) {
              z->next_child = z->mode->submode1->child_cycle->next_child;
              z->mode->submode1->child_cycle->next_child = z;
              z->mode->submode1->child_cycle = z;
            }
            else z->mode->submode1->child_cycle = z->next_child = z;
          }
          else output("FOUND similar mode_mu.\n"); 
*/

          if(back_dfs(z) == TRUE) {
            return TRUE;
          }
        }
        else
          output("In-trans-mode is inconsistent with mu_before.\n");

        if(f==M->in_trans_cycle) break;
      } /* end for(n = ...) */
    }

    output("NO path to initial modes.\n");
    return FALSE;
  }
}

int mode_mu_comp(struct mode_mu_type *mm1, struct mode_mu_type *mm2) {
  int comp;

  comp = mode_comp(mm1->mode, mm2->mode);
  if(comp != EQUAL) return comp;

  comp = bdd_comp(mm1->mu_prp, mm2->mu_prp);
  if(comp != EQUAL) return comp;

  return(zone_comp(zone(mm1->mu_clk), zone(mm2->mu_clk)));
}

int free_mode_mu(struct mode_mu_type *mm) {
  free_mode(mm->mode);
  free(mm);
}

int mode_mu_count_inc(struct mode_mu_type *m) {
  (m->count)++;
}

int mode_mu_parent_set(struct mode_mu_type *m, struct a23tree_type *parent) {
  m->parent = parent;
}

int print_mode_mu(struct mode_mu_type *mm) {
  output("Printing mode_mu:\n");
  output("\tmode: ");
  print_mode(mm->mode);
  output("\tmu_clk: ");
  print_clkconj(mm->mu_clk);
  output("\tmu_prp: ");
  print_bdd(mm->mu_prp);
}

struct clkconj_type *back_delta(struct clkconj_type *c, int ret_type) {
  struct clkconj_type *x, *tmp;
  struct variable_type *v;

  if(c) {
    x = c->next;
    for(; ; x=x->next) {
      if(x->clk2 == 0)
        x->clk2 = DELTA;
      if(x->clk1 == 0)
        x->clk1 = DELTA;
      if(x == c) break;
    }
    x = alloc_clkconj();
    x->clk1 = DELTA;       /* DELTA LEQ 0 */
    if(strict_semantics || ret_type == FALSE)
      x->op = LEQ;
    else
      x->op = LESS;
    tmp = clkconj_and(c, x);
    free_clkconj(c);
    c = tmp;
    tmp = NULL;

    for(v=var_tail->next; ; v=v->next) {
      if(v->type == V_CLOCK) {
        x = alloc_clkconj();
        x->clk1 = v->index;
        x->op = GEQ;
        tmp = clkconj_and(c, x);
        free_clkconj(c);
        c = tmp;
        tmp = NULL;
      }
      if(v == var_tail) break;
    }
/*
    output("After back_delta: ");
    print_clkconj(c); output("\n");
*/
  }
  return c;
}

int initial_consistent(struct mode_mu_type *x) {
  int i = 0;
  char *tmp_name = str_preserve(x->mode->name);
  char *tc = (char *)malloc(sizeof(char) * (strlen(tmp_name) + 1)); 

  tc = strtok(tmp_name, ".");
  if(strcmp(tc, thread[0].starting_mode->name)) return FALSE;

  while(tc = strtok((char *)NULL, ".")) {
    if(strcmp(tc, thread[++i].starting_mode->name)) return FALSE;
  }
  if(!consistent_inv_mu(x->mode, x->mu_clk, x->mu_prp)) return FALSE;
  return TRUE;
}

struct clkconj_type *reset_clk_assign(struct clkconj_type *c, 
                                      struct assign_type *a) {
  struct assign_type *aa;

/*
  output("Resetting assignment: "); print_assignment1(a); output("\n");
  output("for: "); print_clkconj(c); output("\n");
*/
  if(a) {
    for(aa = a->next; ; aa = aa->next) {
      if(get_var_type(aa->var) == V_CLOCK) {
        c = conjoin(c, aa->var);
      }
      if(aa == a) break;
    }
  }
/*
  output("After resetting:\n");
  print_clkconj(c);
  output("\n");
*/
  return c;
}

bdd reset_prp_assign(bdd f, struct assign_type *a) {
  struct assign_type *x;
  struct proposition_type *p;
  int i=0;
  bdd assoc[bdd_vars(bddm)+1];
  bdd t = f;

  if(verbose >= 2) {
    output("Resetting assignment (");
    print_assignment1(a);
    output(")\n in BDD: ");
    print_bdd(f);
  }

  p = prop_tail->next;
  if(a) {
    x = a->next;
    for(; ; x=x->next) {
      if(get_var_type(x->var) == V_REGISTER) {
        for(; ; p = p->next) {
          if(x->var == p->var)
            assoc[i++] = p->prop_bdd;
          if(p == prop_tail) break;
        }
      }
      if(x == a) break;
    }
/*
    sprintf(tmpstring, "No. of vars in temp assoc = %d\n", i);
*/
    assoc[i] = (bdd)NULL;
/*
    output(tmpstring);
*/
    bdd_temp_assoc(bddm, assoc, 0);
    t = bdd_exists(bddm, t);
  }
  return t;
}

struct clkconj_type *weakest_clk_precondition(struct transition_type *f,
                                              struct mode_mu_type *x,
                                              struct clkconj_type *mu1) {
              /* here, mu1 = conjoin(back_delta(x->mu_clk, FALSE), DELTA); */
  struct clkconj_type *c = clkconj_copy(mu1);
  struct clkconj_type *tmp_clkconj;

  c = reset_clk_assign(c, f->assign_cycle);
  tmp_clkconj = clkconj_and(c, f->clk_trigger_cycle);
  free_clkconj(c);
  c = tmp_clkconj;
  tmp_clkconj = clkconj_and(c, f->mode->xconj_cycle);
  free_clkconj(c);
  c = tmp_clkconj;
  tmp_clkconj = NULL; 
  return c;
}
 
bdd weakest_prp_precondition(struct transition_type *f, 
                             struct mode_mu_type *x) {
  bdd b = x->mu_prp;
  b = reset_prp_assign(b, f->assign_cycle);
  b = BDD_AND(b, f->trigger_bdd);
  b = BDD_AND(b, f->mode->boolexp->boolexp_bdd);
  return b;
}

